/*
 * Program to get the host certificate directly from the server using
 * native protocols.
 *
 * Compile with something like:
 *
 *    gcc -I/usr/kerberos/include -lssl -Wall -pedantic get-cert.c -o get-cert
 */

#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <libgen.h> /* for basename */
#include <signal.h>
#include <openssl/bio.h>
#include <openssl/ssl.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/x509.h>

#define BUF_SZ 1111 /* for base64 bug in SL3 openssl0.9.7a */
#define DEFAULT_TIMEOUT 60

int verbose = 0;
int timeout = DEFAULT_TIMEOUT;

/* supported types */
enum proto_type {
	PROTO_DEFAULT = 0,
	PROTO_SSL = 0,
	PROTO_LFC,
	PROTO_GSIFTP,
	PROTO_COUNT
} type = PROTO_DEFAULT; /* default type */

/* protocol handler declaration */
int proto_ssl(BIO*, BIO*, SSL*);
int proto_lfc(BIO*, BIO*, SSL*);
int proto_gsiftp(BIO*, BIO*, SSL*);

/* type details */
struct proto {
	char* name;
	char* default_port;
	int (*handler)(BIO*, BIO*, SSL*);
} proto[] = {
	{ "ssl", "8443", proto_ssl },
	{ "lfc", "5010", proto_lfc },
	{ "gsiftp", "2811", proto_gsiftp }
};

void dump_hex(char buf[], int size, FILE* file) {
	int size16, i;
	char a[17];

	a[16] = '\0';

	size16 = ((size + 15) / 16) * 16; /* round up to 16 */
	for (i = 0; i < size16; i++) {
		if (i % 16 == 0)
			fprintf(file, "%08x  ", i);
		
		if (i >= size)
			fprintf(file, "   ");
		else
			fprintf(file, "%02x ", (unsigned char) buf[i]);
		
		if (i >= size)
			a[i % 16] = ' ';
		else if (isprint(buf[i]))
			a[i % 16] = buf[i];
		else
			a[i % 16] = '.';
		
		if ((i + 1) % 8 == 0)
			fprintf(file, " ");

		if ((i + 1) % 16 == 0)
			fprintf(file, "|%s|\n", a);
	}
}

void verbose_print(int level, char *msg) {
	if (verbose >= level)
		fputs(msg, stderr);
}

void verbose_dump(int level, char *buf, int size) {
	if (verbose >= level)
		dump_hex(buf, size, stderr);
}

void timeout_handler(int signal) {
	char msg[BUF_SZ];
	sprintf(msg, "Timeout after %d seconds\n", timeout);
	verbose_print(0, msg);
	exit(1);
}

int get_proto_from_url(char* url) {
	int i, len;

	for (i = 0; i < PROTO_COUNT; i++) {
		len = strlen(proto[i].name);

		if (strncmp(url, proto[i].name, len) == 0)
			return i;
	}

	return PROTO_DEFAULT;
}

char* get_host_port(int argc, char* argv[]) {
	char *port, *url, *col;
	char *hostname = NULL;
	char *port_ptr = NULL;

	if (argc == 1) /* sanity check */
		return NULL;
		
	/* parse url like proto://host.name:port/path */
	url = argv[1];
	col = strchr(url, ':');

	if (col && col[1] == '/' && col[2] == '/') { /* skip proto:// */
		if (type == PROTO_DEFAULT) /* not assigned explicitly */
			type = get_proto_from_url(url);

		url = col + 3;
		col = strchr(url, ':');
	}

	if (!col) { /* hostname/path */
		char *path = strchr(url, '/');
		if (path)
			hostname = strndup(url, path - url);
		else
			hostname = strdup(url);
	}

	/* get port */
	port = proto[type].default_port;
	if (col && isdigit(col[1])) { /* from url */
		port_ptr = col + 1;
		hostname = strndup(url, col - url);
	}
	if (argc > 2) /* from second argument */
		port_ptr = argv[2];
	if (port_ptr) {
		char* c = port_ptr;
		
		while(isdigit(*c))
			c++;
		if (c - port_ptr > 0) /* got some digits */
			port = strndup(port_ptr, c - port_ptr);
	}

	if (!hostname || !port) /* could not find anything */
		return NULL;

	/* merge all together */
 	hostname = realloc(hostname, strlen(hostname) + 1 + strlen(port) + 1);
	strcat(hostname, ":");
	strcat(hostname, port);

	return hostname;
}

void usage(char* prog) {
	fprintf(stderr, "\
usage: %s [-fhlsv] host.name [port]\n\
\n\
options:\n\
 -f use gsiftp protocol\n\
 -h this help\n\
 -l use lcf protocol\n\
 -s use ssl protocol\n\
 -t timeout (default is %d seconds)\n\
 -v verbose\n", basename(prog), DEFAULT_TIMEOUT);
}

void parse_args(int argc, char* argv[]) {
	int opt;

	while (1) {
		opt = getopt(argc, argv, "fhlst:v");

		if (opt == -1)
			break;

		switch (opt) {
			case 'f':
				type = PROTO_GSIFTP;
				break;
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'l':
				type = PROTO_LFC;
				break;
			case 's':
				type = PROTO_SSL;
				break;
			case 't':
				timeout = atoi(optarg);
				break;
			case 'v':
				verbose++;
				break;
			case '?':
				usage(argv[0]);
				exit(1);
		}
	}
}

int proto_gsiftp(BIO* pipe_bio, BIO* srv_bio, SSL* ssl) {
	BIO *b64_bio, *mem_bio;
	char buf[BUF_SZ];
	char *init = "AUTH GSSAPI\n";
	int n, skip;

	/* chain for base64 enc/dec: base64 <-> mem */
	b64_bio = BIO_new(BIO_f_base64());
	BIO_set_flags(b64_bio, BIO_FLAGS_BASE64_NO_NL);
	mem_bio = BIO_new(BIO_s_mem());
	BIO_push(b64_bio, mem_bio);

	verbose_print(1, "Receiving greeting\n");
	if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
		return 0;
	verbose_dump(2, buf, n);

	verbose_print(1, "Sending auth init\n");
	verbose_dump(2, init, n = strlen(init));
	if (BIO_write(srv_bio, init, n) != n)
		return 0;

	verbose_print(1, "Receiving response\n");
	if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
		return 0;
	verbose_dump(2, buf, n);

	verbose_print(1, "Reading ClientHello\n");
	if ((n = BIO_read(pipe_bio, buf, sizeof(buf))) <= 0)
		return 0;
	verbose_dump(2, buf, n);
	
	verbose_print(1, "Encoding ClientHello\n");
	if (BIO_write(b64_bio, buf, n) != n)
		return 0;
	if (BIO_flush(b64_bio) <= 0)
		return 0;
	strncpy(buf, "ADAT ", sizeof(buf));
	skip = strlen("ADAT ");
	if ((n = BIO_read(mem_bio, buf + skip, sizeof(buf) - skip)) <= 0)
		return 0;
	n += skip;
	buf[n++] = '\n';

	verbose_print(1, "Sending encoded ClientHello\n");
	verbose_dump(2, buf, n);
	if (BIO_write(srv_bio, buf, n) != n)
		return 0;

	skip = strlen("335 ADAT=");
	while (!SSL_get_peer_certificate(ssl)) {
		verbose_print(1, "Receiving encoded ServerHello\n");
		if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
			return 0;
		verbose_dump(2, buf, n);

		verbose_print(1, "Decoding ServerHello\n");
		if (BIO_write(mem_bio, buf + skip, n - skip) != n - skip)
			return 0;
		if ((n = BIO_read(b64_bio, buf, sizeof(buf))) <= 0)
			return 0;
		
		verbose_print(1, "Writing ServerHello\n");
		verbose_dump(2, buf, n);
		if ((BIO_write(pipe_bio, buf, n) != n))
			return 0;

		skip = 0; /* skip header only once (at the beginning) */

		SSL_connect(ssl);
	}

	return 1;
}

int proto_lfc(BIO* pipe_bio, BIO* srv_bio, SSL* ssl) {
	char init[] = { '\x00', '\x00', '\xca', '\x03', /* magic */
			'\x00', '\x00', '\x00', '\x01', /* version */
			'\x00', '\x00', '\x00', '\x23', /* size */
			'\x00', '\x00', '\x00', '\x01', /* request */
			'\x00', '\x00', '\x00', '\x00', /* authid flag */
			'\x00', '\x00', '\x00', '\x02', /* proto count */
			'G',    'S',    'I',    '\x00',
			'I',    'D',    '\x00',
			'\x00', '\x00', '\x00', '\x01', /* flags */
			'\x00', '\x00', '\x00', '\x02',
			'\x00', '\x00', '\x00', '\x01',
			'\x00', '\x00', '\x00', '\x01' };

	char header[] =	{ '\x00', '\x00', '\xca', '\x03', /* magic */
			  '\x00', '\x00', '\x00', '\x01', /* version */
			  '\x00', '\x00', '\x00', '\x00' }; /* size */

	char *c;
	char buf[BUF_SZ];
	int n, skip;
	
	verbose_print(1, "Sending LFC request\n");
	verbose_dump(2, init, sizeof(init));
	if (BIO_write(srv_bio, init, sizeof(init)) != sizeof(init))
		return 0;

	verbose_print(1, "Receiving LFC response\n");
	if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
		return 0;
	verbose_dump(2, buf, n);

	verbose_print(1, "Checking LFC response\n");
	if (!(c = memchr(buf, 'O', n)) || c[1] != 'K')
		return 0;

	verbose_print(1, "Reading ClientHello\n");
	if ((n = BIO_read(pipe_bio, buf + sizeof(header),
			sizeof(buf) - sizeof(header))) <= 0)
		return 0;
	verbose_dump(2, buf + sizeof(header), n);
	
	/* add header */
	memcpy(buf, header, sizeof(header));
	buf[sizeof(header) - 4] = 0xff & (n >> 24);
	buf[sizeof(header) - 3] = 0xff & (n >> 16);
	buf[sizeof(header) - 2] = 0xff & (n >> 8);
	buf[sizeof(header) - 1] = 0xff & (n);

	verbose_print(1, "Sending wrapped ClientHello\n");
	verbose_dump(2, buf, sizeof(header) + n);
	if (BIO_write(srv_bio, buf, sizeof(header) + n) != sizeof(header) + n)
		return 0;

	skip = sizeof(header);
	while (!SSL_get_peer_certificate(ssl)) {
		verbose_print(1, "Receiving wrapped ServerHello\n");
		if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
			return 0;
		verbose_dump(2, buf, n);

		verbose_print(1, "Writing ServerHello\n");
		verbose_dump(2, buf + skip, n - skip);
		if ((BIO_write(pipe_bio, buf + skip, n - skip) != n - skip))
			return 0;

		skip = 0; /* skip header only once (at the beginning) */

		SSL_connect(ssl);
	}

	return 1;
}

int proto_ssl(BIO* pipe_bio, BIO* srv_bio, SSL* ssl) {
	char buf[BUF_SZ];
	int n;

	verbose_print(1, "Reading ClientHello\n");
	if ((n = BIO_read(pipe_bio, buf, sizeof(buf))) == 0)
		return 0;
	verbose_dump(2, buf, n);

	verbose_print(1, "Sending ClientHello\n");
	if (BIO_write(srv_bio, buf, n) != n)
		return 0;

	while (!SSL_get_peer_certificate(ssl)) {
		verbose_print(1, "Receiving ServerHello\n");
		if ((n = BIO_read(srv_bio, buf, sizeof(buf))) <= 0)
			return 0;
		verbose_dump(2, buf, n);

		verbose_print(1, "Writing ServerHello\n");
		if (BIO_write(pipe_bio, buf, n) != n)
			return 0;

		SSL_connect(ssl); /* try to resume */
	}
	
	return 1;
}

int main(int argc, char* argv[]) {
	BIO *srv_bio, *ssl_bio, *pipe_bio;
	SSL_CTX *context;
	SSL *ssl;
	X509 *cert;
	char *server;

	parse_args(argc, argv);
	server = get_host_port(argc - optind + 1, argv + optind - 1);
	if (!server) {
		verbose_print(0, "No hostname specified\n");
		usage(argv[0]);
		exit(1);
	}

	/* setup timeout */
	if (timeout > 0) {
		signal(SIGALRM, &timeout_handler);
		alarm(timeout);
	}

	/* ssl init */
	SSL_load_error_strings();
	SSL_library_init();
	context = SSL_CTX_new(SSLv3_client_method());
	ssl = SSL_new(context);

	/* chain for ssl messages: ssl <-> pipe */
	BIO_new_bio_pair(&ssl_bio, 0, &pipe_bio, 0);
	SSL_set_bio(ssl, ssl_bio, ssl_bio);
	SSL_connect(ssl); /* generate ClientHello */

	/* connect to server */
	verbose_print(1, "Connecting to: ");
	verbose_print(1, server);
	verbose_print(1, "\n");
	srv_bio = BIO_new_connect(server);
	if (BIO_do_connect(srv_bio) <= 0) {
		verbose_print(0, (char *)ERR_reason_error_string(
			ERR_get_error()));
		verbose_print(0, "\n");
		exit(1);
	}

	/* do the protocol */
	verbose_print(1, "Using protocol: ");
	verbose_print(1, proto[type].name);
	verbose_print(1, "\n");
	if (!proto[type].handler(pipe_bio, srv_bio, ssl)) {
		verbose_print(0, "Protocol failed\n");
		exit(1);
	}

	/* close the connection to server */
	BIO_shutdown_wr(srv_bio);

	/* print host cert as pem */
	cert = SSL_get_peer_certificate(ssl);
	if (cert) {
		PEM_write_X509(stdout, cert);
		return 0;
	}

	return 1;
}
